# -------------------------------------------------------------------------------- #
# Ingest Module
# -------------------------------------------------------------------------------- #

# -------------------------------------------------------------------------------- #
# Imports
# -------------------------------------------------------------------------------- #

# Built-in imports
import time
from typing import Tuple, List, Optional

# Logger imports
from src.logger import logger


# OpenAI imports
from openai import AsyncOpenAI

# Environment imports
from dotenv import load_dotenv

# Prompt imports
from src.astral_ai.prompts import GPT4O_VIDEO_INGESTION_SYSTEM_PROMPT, GPT4O_VIDEO_INGESTION_USER_PROMPT,  O1_VIDEO_INGESTION_DEV_PROMPT, O1_VIDEO_INGESTION_USER_PROMPT

# AI imports
from src.models.ai import Model, ReasoningEffort, AICost, AIUsage
from src.astral_ai.exceptions import ImageIngestionNotSupportedError

# Constants imports
from src.constants import REASONING_EFFORT_DEFAULT

# Slide Structured Response imports
from src.models.video import VideoSegment, VideoStructuredResponse, VideoSynthesisRequest, VideoSynthesisStructuredResponse

# Exceptions imports
from src.exceptions import AgenticVideoIngestionFailedError

# Helpers imports
from ai_framework.utils.helpers import track_model_cost, format_video_additional_context, format_user_messages

# Load environment variables
load_dotenv()

# Initialize Async OpenAI client
client = AsyncOpenAI()


# -------------------------------------------------------------------------------- #
# O1 Call with Structured Outputs
# -------------------------------------------------------------------------------- #


async def o1_call(transcript: str,
                  image_urls: List[str],
                  additional_context: str = None,
                  running_context: str = None,
                  reasoning_effort: ReasoningEffort = REASONING_EFFORT_DEFAULT,
                  user_email: str = None) -> Tuple[VideoStructuredResponse, AIUsage, AICost]:
    """
    Call the o1 model to generate cleaned content. 

    Will be used to generate the topic, summary, and content.
    """

    # Format The Developer Prompt
    developer_prompt = O1_VIDEO_INGESTION_DEV_PROMPT + format_video_additional_context(additional_context)

    # Format The User Messages
    user_messages = format_user_messages(transcript=transcript, user_prompt=O1_VIDEO_INGESTION_USER_PROMPT, image_urls=image_urls, running_context=running_context)

    # Log the call
    logger.info(f"Calling {Model.GPT_O1_MODEL} now {' with user-specified additional context' if additional_context else 'without additional context'} and reasoning effort: {reasoning_effort.value}.")
   
    # Format the messages
    messages = [{
        "role": "developer",
        "content": [{
            "type": "text",
            "text": developer_prompt
        }]
    }, {
        "role": "user",
        "content": user_messages
    }]

    # Start timer
    start_time = time.monotonic()

    # Call the model
    response = await client.beta.chat.completions.parse(
        model=Model.GPT_O1_MODEL.value, messages=messages, response_format=VideoStructuredResponse, reasoning_effort=reasoning_effort.value, store=True, user=user_email)

    # End timer
    latency = time.monotonic() - start_time

    # Log finish call with latency
    logger.info(
        f"Finished calling {Model.GPT_O1_MODEL} model. Took: {latency:.2f}s to successfully parse the video content into the anticipated JSON schema."
    )

    # Compute usage and cost
    usage = AIUsage(prompt_tokens=response.usage.prompt_tokens, cached_prompt_tokens=response.usage.prompt_tokens_details.cached_tokens, completion_tokens=response.usage.completion_tokens, total_tokens=response.usage.total_tokens, latency=latency)
    cost = track_model_cost(Model.GPT_O1_MODEL, usage)

    # Return the response
    return response.choices[0].message.parsed, usage, cost


# -------------------------------------------------------------------------------- #
# Parser Call
# -------------------------------------------------------------------------------- #

async def gpt_4o_call(transcript: str,
                      image_urls: List[str],
                      user_email: str,
                      additional_context: Optional[str] = None,
                      running_context: Optional[str] = None) -> Tuple[VideoStructuredResponse, AIUsage, AICost]:
    """
    Call the 4o parser model to parse the video content into the anticipated JSON schema.
    """

    # Format The System Prompt
    system_prompt = GPT4O_VIDEO_INGESTION_SYSTEM_PROMPT + format_video_additional_context(additional_context)

    # Format the user messages
    user_messages = format_user_messages(transcript=transcript, user_prompt=GPT4O_VIDEO_INGESTION_USER_PROMPT, image_urls=image_urls, running_context=running_context)

    # Log the call
    logger.info(f"Calling {Model.GPT_4o_MODEL} now {' with user-specified additional context' if additional_context else 'without additional context'}.")

    # Create the messages
    messages = [{
        "role": "system",
        "content": [{
            "type": "text",
            "text": system_prompt
        }]
    }, {
        "role": "user",
        "content": user_messages
    }]

    # Start timer
    start_time = time.monotonic()

    # Call the model
    response = await client.beta.chat.completions.parse(
        model=Model.GPT_4o_MODEL.value, messages=messages, response_format=VideoStructuredResponse, user=user_email)

    # End timer
    latency = time.monotonic() - start_time

    # Log finish call with latency
    logger.info(
        f"Finished calling {Model.GPT_4o_MODEL} model. Took: {latency:.2f}s to successfully parse the video content into the anticipated JSON schema."
    )

    # Compute usage and cost
    usage = AIUsage(prompt_tokens=response.usage.prompt_tokens, cached_prompt_tokens=response.usage.prompt_tokens_details.cached_tokens, completion_tokens=response.usage.completion_tokens, total_tokens=response.usage.total_tokens, latency=latency)
    cost = track_model_cost(Model.GPT_4o_MODEL, usage)

    # Return the response
    return response.choices[0].message.parsed, usage, cost

# -------------------------------------------------------------------------------- #
# Ingest Main Function
# -------------------------------------------------------------------------------- #


async def agentic_video_ingestion(segment: VideoSegment,
                                  model_name: Model,
                                  reasoning_effort: ReasoningEffort = REASONING_EFFORT_DEFAULT,
                                  additional_context: str = None,
                                  running_context: str = None,
                                  user_email: str = None) -> Tuple[VideoStructuredResponse, AIUsage, AICost]:
    """
    Main function to ingest the video and return:
      (topic, summary, content, usage, cost)

    """

    # Log the call
    logger.info(
        f"Beginning ingestion with the following ingestion model: {model_name.value}. "
        f"{'Additional context was specified.' if additional_context else ''}"
        f"{f' Running context was specified.' if running_context else ''}"
        f"{f' Reasoning effort: {reasoning_effort.value}.' if model_name == Model.GPT_O1_MODEL else ''}"
    )

    # ----------------------------------------------------------
    # If the model is O1-Mini
    # ----------------------------------------------------------
    if model_name == Model.GPT_O1_MINI_MODEL:
        raise ImageIngestionNotSupportedError(model_name)

    # ----------------------------------------------------------
    # If the model is GPT-O1
    # ----------------------------------------------------------

    if model_name == Model.GPT_O1_MODEL:
        # o1_call returns (Content, AIUsage, cost)
        ai_response, ai_usage, ai_cost = await o1_call(
            transcript=segment.transcript,
            image_urls=segment.image_urls,
            reasoning_effort=reasoning_effort,
            additional_context=additional_context,
            running_context=running_context,
            user_email=user_email
        )

    # ----------------------------------------------------------
    # If the model is GPT-4o
    # ----------------------------------------------------------

    if model_name == Model.GPT_4o_MODEL:
        # gpt_4o_call returns (Content, AIUsage, cost)
        ai_response, ai_usage, ai_cost = await gpt_4o_call(
            transcript=segment.transcript,
            image_urls=segment.image_urls,
            additional_context=additional_context,
            running_context=running_context,
            user_email=user_email
        )

    # ----------------------------------------------------------
    # Check if the response is a VideoStructuredResponse
    if not isinstance(ai_response, VideoStructuredResponse):
        raise AgenticVideoIngestionFailedError(f"Failed to ingest video with model: {model_name}. Response: {ai_response}")

    # Return the response
    return ai_response, ai_usage, ai_cost




    # ----------------------------------------------------------




